/***************************************************************************
                          bildstapel.cpp  -  description
                             -------------------
    begin                : Thu Jul 20 2000
    copyright            : (C) 2000 by Immi
    email                : cuyo@pcpool.mathematik.uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstdio>
#include <cstdlib>

#include "inkompatibel.h"
#include "cuyointl.h"

#include "bildstapel.h"
#include "leveldaten.h"
#include "fehler.h"



/** Erzeugt einen uninitialisierten Bildstapel */
BildStapel::BildStapel(bool): mMaxAnz(-1), mDat(0), mPos(0), mViertel(0) {
}


BildStapel::BildStapel(): mSorte(0),
  mMaxAnz(ld->mStapelHoehe), mAnz(-1), 
  mDebugOut1(spezvar_out_nichts), mDebugOut2(spezvar_out_nichts),
  mAmPlatzen(0)
{
  deepCreate();
}


BildStapel::~BildStapel() {
  deepLoesch();
}


BildStapel & BildStapel::operator=(const BildStapel & b) {
  CASSERT(b.mMaxAnz == ld->mStapelHoehe);
  
  /* Stapelhhe ndert sich? Dann Speicher neu alloziieren. */
  if (b.mMaxAnz != mMaxAnz) {
    deepLoesch();
    mMaxAnz = b.mMaxAnz;
    deepCreate();
  }
  kopiere(b);
  return *this;
}


void BildStapel::kopiere(const BildStapel & b) {
  mSorte = b.mSorte;
  //Das wird eh schon oben getan: mMaxAnz = b.mMaxAnz;
  mAnz = b.mAnz;


  for (int i = 0; i < mAnz; i++) {
    mDat[i] = b.mDat[i];
    mPos[i] = b.mPos[i];
    mViertel[i] = b.mViertel[i];
  }
  mAmPlatzen = b.mAmPlatzen;
  mDebugOut1 = b.mDebugOut1;
  mDebugOut2 = b.mDebugOut2;
}



void BildStapel::deepCreate() {
  mDat = new int[mMaxAnz]; CASSERT(mDat);
  mPos = new int[mMaxAnz]; CASSERT(mPos);
  mViertel = new int[mMaxAnz]; CASSERT(mViertel);
}



void BildStapel::deepLoesch() {
  if (mDat)
    delete[] mDat;
  if (mPos)
    delete[] mPos;
  if (mViertel)
    delete[] mViertel;
  mDat = mPos = mViertel = 0;
}



bool BildStapel::operator==(const BildStapel & b) const {
  CASSERT(mDat && mPos && mViertel);
  if (mSorte != b.mSorte)
    return false;
  if (mAnz != b.mAnz)
    return false;
  for (int i = 0; i < mAnz; i++) {
    if (mDat[i] != b.mDat[i]) return false;
    if (mPos[i] != b.mPos[i]) return false;
    if (mViertel[i] != b.mViertel[i]) return false;
  }
  if (mAmPlatzen != b.mAmPlatzen) return false;
  if (mDebugOut1 != b.mDebugOut1) return false;
  if (mDebugOut2 != b.mDebugOut2) return false;
  return true;
}


/** Entfernt alle Bilder und setzt die Sorte. Aufrufen, bevor
    speichereBild() fr jedes Bild aufgerufen wird. */
void BildStapel::initStapel(const Sorte * so, int platz) {
  CASSERT(mDat && mPos && mViertel);
  mSorte = so;
  mAmPlatzen = platz;
  mAnz = 0;
}


void BildStapel::speichereBild(int dat, int pos, int viertel) {
  CASSERT(mDat && mPos && mViertel);
  if (mAnz >= mMaxAnz)
    throw iFehler(_("Internal error: Too many picts drawn for one single blob"));

  /* Gleich mal testen, ob die Werte fr dat, pos und viertel ok sind. */
  /* Throwt, wenn dat out of range ist. */
  int maxpos = mSorte->getBilddatei(dat)->anzBildchen();
  if (pos < 0 || pos >= maxpos)
    throw Fehler(_("Position pos=%d out of range (allowed for file=%d: 0 - %d)"),
                 pos, dat, maxpos - 1);

  if (viertel < viertel_min || viertel > viertel_max)
    throw Fehler(_("Quarter qu=%d out of range (allowed: %d - %d)"),
                 viertel, viertel_min, viertel_max);

  mDat[mAnz] = dat;
  mPos[mAnz] = pos;
  mViertel[mAnz] = viertel;
  mAnz++;
}


void BildStapel::setDebugOut(int d1, int d2) {
  CASSERT(mDat && mPos && mViertel);
  mDebugOut1 = d1;
  mDebugOut2 = d2;
}


/** malt den Bildstapel. xx und yy sind in Pixeln angegeben;
    der Hintergrund wird vorher gelscht. */
void BildStapel::malen(QPainter & p, int xx, int yy) const {
  CASSERT(mDat && mPos && mViertel);
  /* ggf. Position senkrecht spiegeln */
  if (ld->mSpiegeln) yy = gry * gric - yy - gric;


  /* Unter gewissen Umstnden (wenn nicht gemalt werden darf) wird mAnz
     auf was negatives gesetzt. Nur um diesen Sicherheitscheck machen
     zu knnen. */
  CASSERT(mAnz >= 0);

  /* Wenn wirklich was gemalt wird, muss auch eine Sorte angegeben sein. */
  CASSERT(mAnz == 0 || mSorte);

  /* Bild-Stapel malen. */
  for (int i = 0; i < mAnz; i++) {

    try { // Um die Fehlermeldung zu verbessern...
      mSorte->getBilddatei(mDat[i])->
         malBildchen(p, xx, yy, mPos[i], mViertel[i]);
    } catch (Fehler fe) {
      throw Fehler(_("In file %d:\n%s"), mDat[i], fe.getText().data());
    }
  }

  /* Evtl. Explosion malen */
  if (mAmPlatzen) {
    ld->mExplosionBild.malBildchen(p, xx, yy, mAmPlatzen - 1);
  }

  /* Debug-Output vom Animations-Code? */
  if (mDebugOut1 != spezvar_out_nichts)
    malDebug(p, xx, yy, mDebugOut1);
  if (mDebugOut2 != spezvar_out_nichts)
    malDebug(p, xx, yy + gric / 2, mDebugOut2);
} // malen



/** Liefert true, wenn der Stapel leer ist. Wird benutzt, um
    eine Fehlermeldung auszuspucken, wenn whrend eines Events
     gemalt wird. */
bool BildStapel::istLeer() const {
  CASSERT(mDat && mPos && mViertel);
  return mAnz == 0;
}



/** Gibt die Zahl n aus. Wird fr Debug-Output von malen()
    benutzt. */
void BildStapel::malDebug(QPainter & p, int xx, int yy, int n) const {
  CASSERT(mDat && mPos && mViertel);
  __String s;
  s.sprintf("%d", n);
  for (int i = 0; i < (int) s.length(); i++) {
    int z = s[i] == '-'                ? 10 :          // 10 = "-"
            s[i] >= '0' && s[i] <= '9' ? s[i] - '0' :
            11;                                      // 11 = nix
    ld->mDbgZiffernBild.malZiffer(p, xx, yy, z);
    xx += gric/4;    
  }
}



